package article_test

import (
	"context"
	"fmt"
	"math/rand"
	"sync"
	"testing"
	"time"
)

type App struct {
	counter int
	mu      sync.Mutex
	cond    *sync.Cond
}

func NewApp() *App {
	var app App
	app.cond = sync.NewCond(&app.mu)
	return &app
}

func (app *App) Start(ctx context.Context) {
	go app.cleanupWorker(ctx)
}

func (app *App) Increment(id int) {
	// id intは出力を少しわかりやすくするために本誌コードと少しだけ
	// 変えてあります。
	app.mu.Lock()
	defer app.mu.Unlock()
	app.counter++
	fmt.Println("goroutine ", id, ": incremented to ", app.counter)
	app.cond.Signal()
}

func (app *App) cleanupWorker(ctx context.Context) {
	for {
		app.cond.L.Lock()
		for app.counter < 10 {
			app.cond.Wait()
		}
		fmt.Println("counter exceeded 10: ", app.counter)
		app.counter = 0
		app.mu.Unlock()

		fmt.Println("performing cleanup ....")
	}
}

func TestApp(t *testing.T) {
	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	app := NewApp()

	app.Start(ctx)

	for i := 0; i < 5; i++ {
		id := i
		go func(ctx context.Context, id int) {
			for {
				delay := time.Duration(rand.Int31n(int32(time.Second)))
				fmt.Println("goroutine ", id, ": waiting for ", delay)
				timer := time.NewTimer(delay)
				select {
				case <-ctx.Done():
					return
				case <-timer.C:
				}

				app.Increment(id)
			}
		}(ctx, id)
	}

	select {
	case <-ctx.Done():
	}

	fmt.Println("main routine exiting")
}

func TestBroadcast(t *testing.T) {
	// ブロードキャストの例では、待っている状態の複数のgoroutineに
	// 一斉通知を送り、それを受けたgoroutineがそれぞれ反応する、
	// というだけのコードを紹介します。

	var mu sync.Mutex
	cond := sync.NewCond(&mu)

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	// ブロードキャストを待つgoroutine
	for i := 0; i < 5; i++ {
		id := i
		go func(ctx context.Context, id int) {
			for {
				select {
				case <-ctx.Done():
					return
				default:
				}

				mu.Lock()
				cond.Wait()
				fmt.Println("goroutine ", id, ": receieved broadcast!")
				mu.Unlock()
			}
		}(ctx, id)
	}

	// ブロードキャストを送るgoroutine
	go func(ctx context.Context) {
		for {
			delay := time.Duration(rand.Int31n(int32(time.Second)))
			timer := time.NewTimer(delay)
			select {
			case <-ctx.Done():
				return
			case <-timer.C:
			}
			fmt.Println("sender: Sending broadcast")
			cond.Broadcast()
		}
	}(ctx)

	select {
	case <-ctx.Done():
	}

	fmt.Println("main routine exiting")
}
